import .Makie: Point2f, linesegments!, cgrad, plot

############################################################################################
## Plot the Simulation
############################################################################################

@Makie.recipe(SimPlot, sim) do scene
    Makie.Theme(

        # Axis = (
        #     aspect = Makie.DataAspect(),
        #     #backgroundcolor=:gray
        # )
    )
    Makie.Attributes(
        #ac = JULIADYNAMICS_COLORS[1],
        #as = 10,
        #am = :circle,
        scatterkwargs = NamedTuple(),
        particles = (
            color    = :blue,
            colormap = :cyclic_mygbm_30_95_c78_n256_s25,),
        obstacles = NamedTuple(),
        domain = NamedTuple(),

    )
end
Makie.plottype(::Simulation) = SimPlot

function Makie.plot!(plot::SimPlot)#{<:Tuple{<:Simulation}})

    sim = plot[:sim] #Makie.lift(get_sim, plot[1])
    pcs = Makie.lift(particles, sim)
    #particle_style = get(plot.attributes.attributes, :particles, NamedTuple())
    Makie.plot!(plot, pcs; plot[:particles]...)
    Makie.lift(sim) do sim
        if hasobstacles(typeof(sim.particles))
            #obstacle_style = get(plot.attributes.attributes, :obstacles, NamedTuple())
            Makie.plot!(plot, obstacles(sim); plot[:obstacles]...)
        end
    end

    #domain = get(plot.attributes.attributes, :domain, NamedTuple())
    domain = Makie.lift(s->s.domain, sim)
    Makie.plot!(plot, domain; plot[:domain]...)
    return simplot
end

############################################################################################
## Plot Domain Boundaries
############################################################################################

struct PlotDomain{W}
    domain::W
end
get_domain(pd::PlotDomain) = pd.domain
Makie.plottype(::AbstractDomain) = Makie.LineSegments
Makie.convert_arguments(::Type{<:Makie.LineSegments}, w::AbstractDomain) = (PlotDomain(w), )

domain_theme = Makie.Theme(
        color      = :green,
        linewidth = 5,
)

function Makie.default_theme(scene,::Type{<:Makie.LineSegments{<:Tuple{PlotDomain}}})
    merge!(
        domain_theme,
        Makie.default_theme(scene, Makie.LineSegments)
    )
end

function Makie.plot!(plot::Makie.LineSegments{<:Tuple{PlotDomain}})
    w = get_domain(plot[1][])#Makie.lift(get_domain, plot[1])
    bcs = getboundaries(w)

    yls = bcs[2] == PeriodicBoundary ? :dash : :solid
    xls = bcs[1] == PeriodicBoundary ? :dash : :solid
    linesegments!(plot, [Point2f(0.0,0.0), Point2f(w.size[1],0.0)];
        linestyle=xls, plot.attributes.attributes...)
    linesegments!(plot, [Point2f(w.size[1],0.0),Point2f(w.size[1], w.size[2])];
        linestyle=yls, plot.attributes.attributes...)
    linesegments!(plot, [Point2f(w.size[1], w.size[2]),Point2f(0.0, w.size[2])];
        linestyle=xls, plot.attributes.attributes...)
    linesegments!(plot, [Point2f(0.0, w.size[2]),Point2f(0.0, 0.0)];
        linestyle=yls, plot.attributes.attributes...)
end


############################################################################################
## Plot Particles
############################################################################################

struct PlotParticles{V}
    pv::V
end
get_particles(pc::PlotParticles) = pc.pv
Makie.plottype(::AbstractVector{<:AbstractParticle}) = Makie.Mesh
Makie.plottype(::AbstractVector{<:AbstractParticle}, ::Any) = Makie.Mesh


function Makie.convert_arguments(::Type{<:Makie.Mesh}, pv::AbstractVector{<:AbstractParticle})
    (PlotParticles(pv), )
end


function Makie.plot!(plot::Makie.Mesh{<:Tuple{PlotParticles}})
    # this bit is broken
    #poly_kwargs = get(plot.attributes.attributes, :polykwargs, NamedTuple())
    color = get(plot.attributes.attributes, :color, :blue)
    colorfun = get(plot.attributes.attributes, :colorfun, nothing)
    pcs = Makie.lift(get_particles, plot[1])

    if !isnothing(colorfun)
        color = colorfun[].(pcs[])
    end
    particle_polygons = get_particle_polygons(pcs[], NamedTuple())

    pp = Makie.poly!(plot, particle_polygons;
        plot.attributes.attributes...,
        color
        )

    Makie.on(pcs) do cs
        particle_polygons = get_particle_polygons(cs, NamedTuple())
        if !isnothing(colorfun)
            color = colorfun[].(cs)
        end
        pp.color.val = color
        pp[1].val = particle_polygons
        pp.color[] = pp.color[]
        pp[1][] = pp[1][]

    end
    return pp
end


function get_particle_polygons(ps::AbstractVector, polykwargs = NamedTuple())
    offset = SA[0, 0]
    polygons = map(p->getpolygon(p; polykwargs...) .+ (offset'), ps)
    pointpolygons = map(polygons) do poly
        pp = map(axes(poly, 1)) do n
            Point2f(poly[n,1], poly[n,2])
        end
    end
    return pointpolygons
end


particles_theme = Makie.Theme(
        shading = false,
        color    = :blue,
        colormap = :cyclic_mygbm_30_95_c78_n256_s25
)

function Makie.default_theme(scene,::Type{<:Makie.Mesh{<:Tuple{PlotParticles}}})
    merge!(
        particles_theme,
        Makie.default_theme(scene, Makie.Mesh)
    )

end

############################################################################################
## Plot Obstacles
############################################################################################

_stdobstaclecolor_makie(::AbstractObstacle) = :grey
_stdobstaclecolor_makie(o::Union{RodObstacle, DiskObstacle}) = o.inverted ? :white : :grey

struct PlotObstacle{O}
    obstacle::O
end
get_obstacle(po::PlotObstacle) = po.obstacle
Makie.plottype(::AbstractVector{<:AbstractObstacle}) = Makie.Mesh

function Makie.convert_arguments(::Type{<:Makie.Mesh}, po::AbstractVector{<:AbstractObstacle})
    (PlotObstacle(po), )
end

function Makie.plot!(plot::Makie.Mesh{<:Tuple{PlotObstacle}})
    @show pos = Makie.lift(get_obstacle, plot[1])
    particle_polygons = Makie.lift(get_particle_polygons, pos, get(plot.attributes.attributes, :polykwargs, NamedTuple()))
    @show particle_polygons[]
    color = plot[:color][]
    if color == :default
        color = _stdobstaclecolor_makie.(pos[])
    end
    println(plot.attributes.attributes)
    println(color)
    Makie.poly!(plot, particle_polygons;
        plot.attributes.attributes...,
        color
        )
end

obstacle_theme = Makie.Theme(
        color      = :default,
        shading = false,
)

function Makie.default_theme(scene,::Type{<:Makie.Mesh{<:Tuple{PlotObstacle}}})
    merge!(
        obstacle_theme,
        Makie.default_theme(scene, Makie.Mesh)
    )

end
